# -*- coding: utf-8 -*-

"""
Term-document Matrix class.
"""

import numpy as np
from typing import Dict, Tuple
from textarium.corpus import Corpus
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from prettytable import PrettyTable

class TermDocMatrix:
    def __init__(self, vectorizer_type: str, ngram_range: Tuple = (1, 1), analyzer: str = 'word'):
        self.type = vectorizer_type
        self.ngram_range = ngram_range
        self.analyzer = analyzer
        self.matrix = None
        self.vectorizer = None
        self.tokens = None

    def from_corpus(self, corpus: Corpus, use_raw : bool = False):
        """Generate term-doc matrix from Corpus texts.

        Args:
            corpus (Corpus): Corpus class object.

        Raises:
            Exception: If provided TermDocMatrix 'type' parameter does not exist.
        """
        if self.type == 'count':
            self.vectorizer = CountVectorizer(ngram_range=self.ngram_range, analyzer=self.analyzer)
        elif self.type == 'tfidf':
            self.vectorizer = TfidfVectorizer(ngram_range=self.ngram_range, analyzer=self.analyzer)
        else:
            raise Exception("There is no such type")
        self.matrix = self.vectorizer.fit_transform(corpus.to_array(use_raw=use_raw))
        self.tokens = self.vectorizer.get_feature_names_out()

    def term_frequency(self, ascending=False, print_n: int = 10) -> Dict:
        """Calculate frequencies for each term.

        Args:
            ascending (bool, optional): Sorting order. Defaults to False.
            print_n (int, optional): Number of top elements to print. Defaults to 10.

        Returns:
            Dict: A sorted dictionary with term frequencies.
        """
        if self.type != 'count':
            raise Exception("Available only for TermDocMatrix generated with CountVectorizer")
        freqs = dict(zip(self.tokens, np.squeeze(np.asarray(self.matrix.sum(axis=0)))))
        sorted_freqs = dict(sorted(freqs.items(), key=lambda item: item[1], reverse=not ascending)[:print_n])
        tab = PrettyTable(["Term", "Count"])
        for k, v in sorted_freqs.items():
            tab.add_row([k, v])
        print(tab)
        return sorted_freqs

    def term_doc_frequency(self, ascending=False, print_n: int = 10) -> Dict:
        """Calculate term-doc frequencies for each term.

        Args:
            ascending (bool, optional): Sorting order. Defaults to False.
            print_n (int, optional): Number of top elements to print. Defaults to 10.

        Returns:
            Dict: A sorted dictionary with term-doc frequencies.
        """
        if self.type != 'count':
            raise Exception("Available only for TermDocMatrix generated with CountVectorizer")
        freqs = dict(zip(self.tokens, np.squeeze(np.asarray(self.matrix.astype(bool).sum(axis=0)))))
        sorted_freqs = dict(sorted(freqs.items(), key=lambda item: item[1], reverse=not ascending)[:print_n])
        tab = PrettyTable(["Term", "Count"])
        for k, v in sorted_freqs.items():
            tab.add_row([k, v])
        print(tab)
        return sorted_freqs



    




# -*- coding: utf-8 -*-

"""
Function for extracting information from texts.
"""

import re
import string
import yake
from typing import List, Tuple
from razdel import sentenize
from nltk import tokenize
from nltk.util import ngrams
from datetime import datetime
from dateutil import parser
from textarium.utils import date_parser_jumpwords, date_parser_keywords
import textarium.preprocessing as preprocessing

def extract_words(text: str) -> List[str]:
    """Extract a list of tokens from a text

    Args:
        text (str): Any string

    Returns:
        List[str]: A list of extracted tokens
    """
    words = tokenize.word_tokenize(text)
    return words

def extract_sentences(text: str, lang='en') -> List[str]:
    """Extract a list of sentences from a text

    Args:
        text (str): Any string in English or Russian
        lang (str): Text language ('ru' - Russian or 'en' - English)

    Returns:
        List[str]: A list of extracted sentences
    """
    if lang == 'en':
        sentences = tokenize.sent_tokenize(text)
        sentences = [preprocessing.remove_extra_spaces(s) for s in sentences]
    elif lang == 'ru':
        sentences = [i.text for i in sentenize(text)]
    return sentences

def extract_urls(text: str) -> List[str]:
    """Extract a list of URLs from a text

    Args:
        text (str): Any string

    Returns:
        List[str]: A list of extracted URLs
    """
    url_regex = re.compile(
        "((https?):((//)|(\\\\))+([\w\d:#@%/;$()~_?\+-=\\\.&](#!)?)*)", re.DOTALL
    )
    parsed_url_objects = re.findall(url_regex, text)
    urls = [url_param[0].strip(string.punctuation) for url_param in parsed_url_objects]
    return urls

def extract_emails(text: str) -> List[str]:
    """Extract a list of E-mails from a text

    Args:
        text (str): Any string

    Returns:
        List[str]: A list of extracted E-mails
    """
    email_regex = re.compile(
        "[a-zA-Z0–9._%+-]+@[a-zA-Z0–9.-]+\.[a-zA-Z]{2,}", re.DOTALL
    )
    parsed_email_objects = re.findall(email_regex, text)
    return parsed_email_objects

def extract_dates(text: str) -> List[datetime]:
    """Extract a list of dates (datetime.datetime)

    Args:
        text (str): Any string

    Returns:
        List[datetime]: A list of dates
    """
    def is_valid_kw(s):
        try:  # is it a number?
            float(s)
            return True
        except ValueError:
            return s.lower() in date_parser_keywords

    def _split(s):
        kw_found = False
        tokens = parser._timelex.split(s)
        for i in range(len(tokens)):
            if tokens[i] in date_parser_jumpwords:
                continue 
            if not kw_found and is_valid_kw(tokens[i]):
                kw_found = True
                start = i
            elif kw_found and not is_valid_kw(tokens[i]):
                kw_found = False
                yield "".join(tokens[start:i])
        if kw_found:
            yield "".join(tokens[start:])
    return [parser.parse(x) for x in _split(text)]

def extract_ngrams(text: str, n: int) -> List[Tuple]:
    """Extract a list of ngrams (tuples of a given length)

    Args:
        text (str): Any string
        n (int): A number of words in ngram

    Returns:
        List[Tuple]: A list of extracted ngrams
    """
    return list(ngrams(extract_words(text), n))

def extract_keywords(text: str, lang: str = 'en', method: str = 'yake') -> List[Tuple]:
    """Extract keywords from a given text using a chosen approach.

    Args:
        text (str): Any string
        lang (str, optional): A language of a text. Defaults to 'en'.
        method (str, optional): Choose an approach to extract keywords: 'yake'. Defaults to 'yake'.

    Returns:
        List[Tuple]: _description_
    """
    if method == 'yake':
        extractor = yake.KeywordExtractor(lan=lang)
        keywords = extractor.extract_keywords(text)
    else:
        raise Exception("Please, choose an approach from any of these: 'yake' (to be continued)")
    return keywords


from unittest import TestCase

from textarium.text import Text

class TestText(TestCase):
    def test_prepare_en_0(self):
        input_text = """
        Hello! My name is Mr.Parker.
        I have a website https://parker.com.
        It has about 5000 visitors per day.
        I track it with a simple html-block like this:
        <div>Google.Analytics</div>
        """
        expected_result = (
            "hello my name is mr parker"\
            " i have a website it ha about visitor per day"\
            " i track it with a simple html block like this"\
            " google analytics"
        )
        text = Text(input_text, lang='en')
        text.prepare()
        self.assertEqual(expected_result, text.prepared_text)

    def test_prepare_ru_0(self):
        input_text = """
        Добрый день! Мое имя мистер Паркер.
        У меня есть веб-сайт https://parker.com.
        Трекер насчитывает 5000 посетителей в день.
        Я слежу за этим с помощью простого html-блока:
        <div>Google.Analytics</div>
        """
        expected_result = (
            "добрый день мой имя мистер паркер"\
            " у я есть веб сайт трекер насчитывать посетитель в день"\
            " я следить за это с помощь простой html блок"\
            " google analytics"
        )
        text = Text(input_text, lang='ru')
        text.prepare()
        self.assertEqual(expected_result, text.prepared_text)


# -*- coding: utf-8 -*-

"""
Corpus class.
"""

import copy
from typing import List, Callable
from prettytable import PrettyTable
from .text import Text

class Corpus:
    def __init__(self, texts: List[str], lang='en'):
        self.lang = lang
        self.corpus = [Text(t, lang=self.lang) for t in texts]
        self.is_prepared = False

    def __getitem__(self, index):
        return self.corpus[int(index)]

    def __len__(self):
        return len(self.corpus)

    def __str__(self) -> str:
        if len(self.corpus) <= 10:
            return "\n\n".join([t.raw_text for t in self.corpus])
        else:
            first_five_texts = "\n\n".join([t.raw_text for t in self.corpus[:5]])
            last_five_texts = "\n\n".join([t.raw_text for t in self.corpus[-5:]])
            return first_five_texts + "\n\n...\n\n" + last_five_texts

    def __repr__(self) -> str:
        return f"textarium.Corpus object containing {len(self.corpus)} textarium.Text objects"

    def to_array(self, use_raw: bool = False) -> List[str]:
        if self.is_prepared and not use_raw:
            return [t.prepared_text for t in self.corpus]
        else:
            return [t.raw_text for t in self.corpus]

    def prepare(
        self, 
        lemmatize: bool = True, 
        get_keywords: bool = True, 
        stopwords: List[str] = None
        ):
        for t in self.corpus:
            t.prepare(lemmatize=lemmatize, get_keywords=get_keywords, stopwords=stopwords)
        self.is_prepared = True

    def info(self):
        """Print information about the corpus and its content.
        """
        attr_counters = dict()
        for obj in self.corpus:
            for attr, value in obj.__dict__.items():
                if attr not in attr_counters:
                    attr_counters[attr] = [1, 1 if value else 0]
                else:
                    attr_counters[attr][0] += 1
                    attr_counters[attr][1] += 1 if value else 0

        print(f"Corpus consists of {len(self)} texts.")
        print(f"There {len(attr_counters.keys())} available attributes of texts:")
        tab = PrettyTable(["Attribute", "Count", "Non-null"])
        for k, v in attr_counters.items():
            tab.add_row([k, v[0], v[1]])
        print(tab)

    def filter(self, condition: Callable, attribute: str, inplace: bool = False):
        """Filter texts in Corpus by provided condition.
        Condition is a predicat 
        Args:
            condition (Callable): Function that takes text as a argument
        and returns True (keep it) or False (remove it from the corpus).
            filter_by (str, optional): Choose what version of texts to use: "raw" or "prepared". Defaults to "raw".
        """
        if inplace:
            self.corpus[:] = [t for t in self.corpus if condition(getattr(t, attribute))]
        else:
            corpus_copy = copy.deepcopy(self)
            corpus_copy.corpus[:] = [t for t in corpus_copy if condition(getattr(t, attribute))]
            return corpus_copy



# -*- coding: utf-8 -*-

"""
Function for extracting information from texts.
"""

from pymorphy2 import MorphAnalyzer
from nltk.stem import WordNetLemmatizer

morph_ru = MorphAnalyzer()
morph_en = WordNetLemmatizer()


"""
Textarium
=====

Provides
  1. Classes and functions to process and analyze text information
  2. Collections of stopwords and other useful dictionaries for NLP tasks

How to use the documentation
----------------------------
Documentation is available in two forms: docstrings provided
with the code, and a loose standing reference guide, available from
`the Textarium homepage <https://6b656b.github.io/textarium/>`_.

See below for further instructions.

Code snippets are indicated by three greater-than signs::

  >>> from textarium import Text
  >>> text = Text("Hello world!", lang='en')
  >>> text.prepare()
  >>> print(text.prepared_text)

Available subpackages
---------------------
extraction
    Basic functions for extracting any kind of tokens from string
models
    NLP models for text preprocessing and embedding
preprocessing
    Functions for text preprocessing
text
    Text class

Viewing documentation using IPython
-----------------------------------

Start IPython and import `textarium`: `import
textarium`.  Then, directly past or use the ``%cpaste`` magic to paste
examples into the shell.  To see which functions are available in `textarium`,
type ``textarium.<TAB>`` (where ``<TAB>`` refers to the TAB key).

"""
from .extraction import extract_words, extract_urls, extract_sentences
from .preprocessing import (
    remove_extra_spaces,
    remove_charset,
    remove_digits,
    remove_html,
    remove_punctuation,
    remove_urls,
    remove_words,
    make_lowercase,
    lemmatize
)
from .text import Text
from .corpus import Corpus
from .termdoc import TermDocMatrix

__all__ = [
    "extract_words",
    "extract_urls",
    "extract_sentences",
    "remove_extra_spaces",
    "remove_charset",
    "remove_digits",
    "remove_html",
    "remove_punctuation",
    "remove_urls",
    "remove_words",
    "make_lowercase",
    "lemmatize",
    "Text",
    "Corpus",
    "TermDocMatrix",
]


# -*- coding: utf-8 -*-

"""
Stopwords.
"""

from os import path
from typing import List

def get_stopwords(lang: str = 'en') -> List[str]:
    file_path = path.join(path.dirname(__file__), f'{lang}_stopwords.txt')
    with open(file_path, 'r') as f:
        stopwords = f.readlines()
    return stopwords


from unittest import TestCase
from textarium.corpus import Corpus

test_texts_list = [
    "Hello! My name is Mr.Parker.",
    "I have a website https://parker.com.",
    "It has about 5000 visitors per day.",
    "I track it with a simple html-block like this:",
    "<div>Google.Analytics</div>",
]

class TestCorpus(TestCase):
    def test_prepare_en_0(self):
        expected_result = [
            "hello my name is mr parker",
            "i have a website",
            "it ha about visitor per day",
            "i track it with a simple html block like this",
            "google analytics",
        ]
        corpus = Corpus(test_texts_list, lang='en')
        corpus.prepare()
        self.assertListEqual(expected_result, [t.prepared_text for t in corpus.corpus])

    def test_filter_raw_attr_inplace_false(self):
        condition = lambda x : len(x.split()) > 5
        expected_result = [
            "It has about 5000 visitors per day.",
            "I track it with a simple html-block like this:",
        ]
        corpus = Corpus(test_texts_list, lang='en')
        result = corpus.filter(condition=condition, attribute="raw_text")
        self.assertListEqual(expected_result, [t.raw_text for t in result.corpus])

    def test_filter_raw_attr_inplace_true(self):
        condition = lambda x : len(x.split()) > 5
        expected_result = [
            "It has about 5000 visitors per day.",
            "I track it with a simple html-block like this:",
        ]
        corpus = Corpus(test_texts_list, lang='en')
        corpus.filter(condition=condition, attribute="raw_text", inplace=True)
        self.assertListEqual(expected_result, [t.raw_text for t in corpus.corpus])

    def test_filter_prepared_attr_inplace_false(self):
        condition = lambda x : "visitor" in x
        expected_result = [
            "it ha about visitor per day",
        ]
        corpus = Corpus(test_texts_list, lang='en')
        corpus.prepare()
        result = corpus.filter(condition=condition, attribute="prepared_text")
        self.assertListEqual(expected_result, [t.prepared_text for t in result.corpus])

    def test_get_item(self):
        corpus = Corpus(test_texts_list, lang='en')
        idx = 2
        expected_result = test_texts_list[2]
        result = corpus[idx].raw_text
        self.assertEqual(expected_result, result)

    def test_len(self):
        corpus = Corpus(test_texts_list, lang='en')
        expected_result = len(test_texts_list)
        result = len(corpus)
        self.assertEqual(expected_result, result)

    def test_info(self):
        corpus = Corpus(test_texts_list, lang='en')
        corpus.info()




from unittest import TestCase
import collections
import textarium

class TestCollections(TestCase):
    def test_no_duplicates_in_textarium__all__(self):
        dups = {k: v for k, v in collections.Counter(textarium.__all__).items() if v > 1}
        self.assertTrue(len(dups) == 0)


# -*- coding: utf-8 -*-

"""
Utility functions
"""

from functools import reduce
from typing import Any, Callable, Sequence
import itertools
from dateutil import parser

def pipeline(
        arg: Any,
        function_pipeline: Sequence[Callable],
) -> Any:
    """A generic Unix-like pipeline

    Args:
        arg (Any): The value you want to pass through a pipeline
        function_pipeline (Sequence[Callable]): An ordered 
        list of functions that comprise your pipeline

    Returns:
        Any: A result of functions pipeline
    """
    return reduce(lambda v, f: f(v), function_pipeline, arg)


date_parser_jumpwords = set(parser.parserinfo.JUMP)
date_parser_keywords = set(kw.lower() for kw in itertools.chain(
    parser.parserinfo.UTCZONE,
    parser.parserinfo.PERTAIN,
    (x for s in parser.parserinfo.WEEKDAYS for x in s),
    (x for s in parser.parserinfo.MONTHS for x in s),
    (x for s in parser.parserinfo.HMS for x in s),
    (x for s in parser.parserinfo.AMPM for x in s),
))